from openquantum_sdk_qiskit.oq_backend import OpenQuantumBackend


def test_rigetti_ankaa_3_backend():
    backend = OpenQuantumBackend(name="Rigetti Ankaa-3")
    assert backend.num_qubits == 84
    ops = backend.target.operation_names
    assert "rx" in ops
    assert "rz" in ops
    assert "iswap" in ops
    assert "measure" in ops
    assert "cz" not in ops
    assert "cx" not in ops


def test_ionq_aria_1_backend():
    backend = OpenQuantumBackend(name="IonQ Aria-1")
    assert backend.num_qubits == 25
    ops = backend.target.operation_names
    assert "gpi" in ops
    assert "gpi2" in ops
    assert "ms" in ops
    assert "measure" in ops
    assert "cx" not in ops


def test_iqm_emerald_backend():
    backend = OpenQuantumBackend(name="IQM Emerald")
    assert backend.num_qubits == 54
    target = backend.target
    ops = target.operation_names
    assert target.num_qubits == 54
    assert "measure" in ops
    assert "cz" in ops
    assert "prx" in ops
    assert target.instruction_supported("cz", (0, 1)) or target.instruction_supported("cz", (1, 0))


def test_iqm_garnet_backend():
    backend = OpenQuantumBackend(name="IQM Garnet")
    assert backend.num_qubits == 20
    target = backend.target
    assert target.num_qubits == 20
    assert "measure" in target.operation_names
    assert "cz" in target.operation_names
    assert "prx" in target.operation_names
    assert target.instruction_supported("cz", (0, 1)) or target.instruction_supported("cz", (1, 0))



# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# Copyright 2024 IonQ, Inc. (www.ionq.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Portions of this file (IonQ gate definitions and equivalences) are derived
# from qiskit-ionq: https://github.com/qiskit-community/qiskit-ionq

"""Build Qiskit Target from backend capabilities."""

from __future__ import annotations

from typing import Any, Dict, Iterable, Tuple


def build_target_from_capabilities(cap: Dict[str, Any]):
    """build Qiskit Target from a capabilities dict.

    Expects:
      - n_qubits: int
      - timing: { dt: float|None, durations: [{op, qubits, value_s}] }
      - native_ops: [{name, arity}]
      - topology: { directed_edges: bool, coupling_map: [[int,int], ...] }
      - noise: { gate_error: [{op, qubits, prob}] }
    """
    from qiskit.transpiler import Target, InstructionProperties
    from qiskit.circuit import Gate
    from qiskit.circuit.library import (
        XGate, YGate, ZGate, HGate, SXGate, SXdgGate, TGate, TdgGate, SGate, SdgGate,
        RXGate, RYGate, RZGate, PhaseGate,
        CXGate, CZGate, SwapGate, iSwapGate, ECRGate,
        CCXGate, CSwapGate, CPhaseGate, CYGate,
        RXXGate, RYYGate, RZZGate,
        Measure
    )

    n_qubits = int(cap.get("n_qubits", 0))
    dt = (cap.get("timing", {}) or {}).get("dt")
    target = Target(num_qubits=n_qubits, dt=dt)

    durations = {}
    for d in (cap.get("timing", {}) or {}).get("durations", []) or []:
        try:
            key = (str(d["op"]).lower(), tuple(int(q) for q in d["qubits"]))
            durations[key] = float(d["value_s"])
        except Exception:
            continue

    errors = {}
    for e in (cap.get("noise", {}) or {}).get("gate_error", []) or []:
        try:
            key = (str(e["op"]).lower(), tuple(int(q) for q in e["qubits"]))
            errors[key] = float(e["prob"])
        except Exception:
            continue

    def props_for(op_name: str, qs: Iterable[int]):
        key = (op_name.lower(), tuple(qs))
        dur_s = durations.get(key)
        dur_dt = (dur_s / dt) if (dur_s is not None and dt) else None
        err = errors.get(key)
        if dur_dt is None and err is None:
            return None
        return InstructionProperties(duration=dur_dt, error=err)

    directed = bool((cap.get("topology", {}) or {}).get("directed_edges", False))
    coupling = [tuple(map(int, pair)) for pair in ((cap.get("topology", {}) or {}).get("coupling_map", []) or [])]

    class PRXGate(Gate):
        """Phased RX gate: R(theta, phi)"""

        def __init__(self, theta, phi):
            super().__init__("prx", 1, [theta, phi])

        def _define(self):
            from qiskit.circuit.library import RZGate, RXGate
            from qiskit.circuit import QuantumCircuit
            qc = QuantumCircuit(1)
            qc.append(RZGate(self.params[1]), [0])
            qc.append(RXGate(self.params[0]), [0])
            qc.append(RZGate(-self.params[1]), [0])
            self.definition = qc

    class GPIGate(Gate):
        """IonQ GPI gate"""

        def __init__(self, phi):
            super().__init__("gpi", 1, [phi])

        def _define(self):
            import numpy as np
            from qiskit.circuit import QuantumCircuit

            phi = self.params[0]

            unitary = np.array([
                [0, np.exp(-1j * 2 * np.pi * phi)],
                [np.exp(1j * 2 * np.pi * phi), 0]
            ])

            qc = QuantumCircuit(1)
            qc.unitary(unitary, [0], label='gpi')
            self.definition = qc

    class GPI2Gate(Gate):
        """IonQ GPI2 gate."""

        def __init__(self, phi):
            super().__init__("gpi2", 1, [phi])

        def _define(self):
            import numpy as np
            from qiskit.circuit import QuantumCircuit

            phi = self.params[0]

            unitary = (1 / np.sqrt(2)) * np.array([
                [1, -1j * np.exp(-1j * 2 * np.pi * phi)],
                [-1j * np.exp(1j * 2 * np.pi * phi), 1]
            ])

            qc = QuantumCircuit(1)
            qc.unitary(unitary, [0], label='gpi2')
            self.definition = qc

    class MSGate(Gate):
        """IonQ Mølmer-Sørensen gate."""

        def __init__(self, phi0, phi1, theta=0.25):
            super().__init__("ms", 2, [phi0, phi1, theta])

        def _define(self):
            import numpy as np
            from qiskit.circuit import QuantumCircuit

            phi0, phi1, theta = self.params

            diag = np.cos(np.pi * theta)
            sin = np.sin(np.pi * theta)

            unitary = np.array([
                [diag, 0, 0, -1j * sin * np.exp(-1j * 2 * np.pi * (phi0 + phi1))],
                [0, diag, -1j * sin * np.exp(-1j * 2 * np.pi * (phi0 - phi1)), 0],
                [0, -1j * sin * np.exp(1j * 2 * np.pi * (phi0 - phi1)), diag, 0],
                [-1j * sin * np.exp(1j * 2 * np.pi * (phi0 + phi1)), 0, 0, diag],
            ])

            qc = QuantumCircuit(2)
            qc.unitary(unitary, [0, 1], label='ms')
            self.definition = qc

    from qiskit.circuit.equivalence_library import SessionEquivalenceLibrary as SELib
    from qiskit.circuit.library import (  # noqa: F811
        RGate, RXGate, RYGate, RZGate, HGate, CXGate, CZGate
    )
    from qiskit.circuit import QuantumCircuit, Parameter
    import numpy as np

    theta = Parameter("theta")
    phi = Parameter("phi")

    # PRX(θ,φ) = RZ(φ) · RX(θ) · RZ(-φ)
    qc_prx_equiv = QuantumCircuit(1)
    qc_prx_equiv.rz(phi, 0)
    qc_prx_equiv.rx(theta, 0)
    qc_prx_equiv.rz(-phi, 0)
    SELib.add_equivalence(PRXGate(theta, phi), qc_prx_equiv)

    # R(θ,φ) = PRX(θ,φ)
    qc_r = QuantumCircuit(1)
    qc_r.append(PRXGate(theta, phi), [0])
    SELib.add_equivalence(RGate(theta, phi), qc_r)

    # RX(θ) = PRX(θ,0)
    qc_rx = QuantumCircuit(1)
    qc_rx.append(PRXGate(theta, 0.0), [0])
    SELib.add_equivalence(RXGate(theta), qc_rx)

    # RY(θ) = PRX(θ,π/2)
    qc_ry = QuantumCircuit(1)
    qc_ry.append(PRXGate(theta, np.pi / 2), [0])
    SELib.add_equivalence(RYGate(theta), qc_ry)

    # IonQ GPI/GPI2/MS equivalences
    # Derived from qiskit-ionq (Copyright 2024 IonQ, Apache 2.0)
    # https://github.com/qiskit-community/qiskit-ionq
    from qiskit.circuit.library import UGate
    lam = Parameter("lam")

    # GPI(φ)
    qc_gpi_equiv = QuantumCircuit(1)
    qc_gpi_equiv.x(0)
    qc_gpi_equiv.rz(4 * np.pi * phi, 0)
    SELib.add_equivalence(GPIGate(phi), qc_gpi_equiv)

    # GPI2(φ)
    qc_gpi2_equiv = QuantumCircuit(1)
    qc_gpi2_equiv.rz(-2 * np.pi * phi, 0)
    qc_gpi2_equiv.rx(np.pi / 2, 0)
    qc_gpi2_equiv.rz(2 * np.pi * phi, 0)
    SELib.add_equivalence(GPI2Gate(phi), qc_gpi2_equiv)

    # U(θ,φ,λ)
    qc_u_equiv = QuantumCircuit(1)
    qc_u_equiv.append(GPI2Gate(1/2 - lam / (2 * np.pi)), [0])
    qc_u_equiv.append(GPIGate(theta / (4 * np.pi) + phi / (4 * np.pi) - lam / (4 * np.pi)), [0])
    qc_u_equiv.append(GPI2Gate(1/2 + phi / (2 * np.pi)), [0])
    SELib.add_equivalence(UGate(theta, phi, lam), qc_u_equiv)

    # CX
    qc_cx_equiv = QuantumCircuit(2)
    qc_cx_equiv.append(GPI2Gate(1/4), [0])
    qc_cx_equiv.append(MSGate(0, 0, 1/4), [0, 1])
    qc_cx_equiv.append(GPI2Gate(1/2), [0])
    qc_cx_equiv.append(GPI2Gate(1/2), [1])
    qc_cx_equiv.append(GPI2Gate(-1/4), [0])
    SELib.add_equivalence(CXGate(), qc_cx_equiv)

    _lib_map = {
        "x": lambda: XGate(),
        "y": lambda: YGate(),
        "z": lambda: ZGate(),
        "h": lambda: HGate(),
        "s": lambda: SGate(),
        "si": lambda: SdgGate(),
        "sdg": lambda: SdgGate(),
        "t": lambda: TGate(),
        "ti": lambda: TdgGate(),
        "tdg": lambda: TdgGate(),
        "v": lambda: SXGate(),
        "vi": lambda: SXdgGate(),
        "sx": lambda: SXGate(),
        "sxdg": lambda: SXdgGate(),

        "rx": lambda: RXGate(0.0),
        "ry": lambda: RYGate(0.0),
        "rz": lambda: RZGate(0.0),
        "phaseshift": lambda: PhaseGate(0.0),
        "p": lambda: PhaseGate(0.0),

        "cx": lambda: CXGate(),
        "cnot": lambda: CXGate(),
        "cz": lambda: CZGate(),
        "cy": lambda: CYGate(),
        "swap": lambda: SwapGate(),
        "iswap": lambda: iSwapGate(),
        "ecr": lambda: ECRGate(),
        "pswap": lambda: SwapGate(),

        "ccnot": lambda: CCXGate(),
        "ccx": lambda: CCXGate(),
        "cswap": lambda: CSwapGate(),
        "cphaseshift": lambda: CPhaseGate(0.0),
        "cp": lambda: CPhaseGate(0.0),

        "rxx": lambda: RXXGate(0.0),
        "xx": lambda: RXXGate(0.0),
        "ms": lambda: MSGate(0.0, 0.0),
        "ryy": lambda: RYYGate(0.0),
        "yy": lambda: RYYGate(0.0),
        "rzz": lambda: RZZGate(0.0),
        "zz": lambda: RZZGate(0.0),

        "measure": lambda: Measure(),
        "measure_ff": lambda: Measure(),

        # custom gates
        "prx": lambda: PRXGate(0.0, 0.0),
        "gpi": lambda: GPIGate(0.0),
        "gpi2": lambda: GPI2Gate(0.0),
    }

    all_ops = cap.get("native_ops", []) or []

    seen_ops = set()
    unique_ops = []
    for op in all_ops:
        name = str(op.get("name"))
        if name not in seen_ops:
            seen_ops.add(name)
            unique_ops.append(op)

    added_qiskit_names = set()

    for op in unique_ops:
        name = str(op.get("name"))
        arity = int(op.get("arity", 1))
        name_l = name.lower()

        try:
            if name_l in _lib_map:
                gate = _lib_map[name_l]()
            else:
                gate = Gate(name, arity, [])
        except Exception:
            gate = Gate(name, arity, [])

        if gate.name in added_qiskit_names:
            continue

        if arity == 1:
            mapping: Dict[Tuple[int], Any] = {}
            for q in range(n_qubits):
                ip = props_for(name, [q])
                mapping[(q,)] = ip
            if mapping:
                target.add_instruction(gate, mapping)
                added_qiskit_names.add(gate.name)
        elif arity == 2:
            mapping2: Dict[Tuple[int, int], Any] = {}
            for a, b in coupling:
                ip = props_for(name, [a, b])
                mapping2[(a, b)] = ip
                if not directed:
                    ip2 = props_for(name, [b, a])
                    mapping2[(b, a)] = ip2
            if mapping2:
                target.add_instruction(gate, mapping2)
                added_qiskit_names.add(gate.name)
    return target



from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional, Dict, Any  # noqa: F401
from openquantum_sdk.qiskit import OpenQuantumService
from openquantum_sdk.models import (
    BackendClassRead,
    PaginatedBackendClasses,
    PaginationInfo,
)


@dataclass
class _Page:
    backend_classes: List[BackendClassRead]
    pagination: PaginationInfo


def _bk(idx: int, **kw) -> BackendClassRead:
    return BackendClassRead(
        id=f"bk-{idx}",
        name=f"Backend {idx}",
        description="",
        type=kw.get("type", "QPU"),
        provider_id=kw.get("provider_id", "prov-1"),
        short_code=kw.get("short_code", f"sc-{idx}"),
        queue_depth=kw.get("queue_depth", 0),
        accepting_jobs=kw.get("accepting_jobs", True),
        status=kw.get("status", "Online"),
    )


def test_service_backends_filters_and_pagination(monkeypatch):
    svc = OpenQuantumService()

    pages = [
        _Page(
            backend_classes=[_bk(1), _bk(2, type="SIMULATOR")],
            pagination=PaginationInfo(next_cursor="cursor-2"),
        ),
        _Page(
            backend_classes=[_bk(3, accepting_jobs=False, status="Offline"), _bk(4)],
            pagination=PaginationInfo(next_cursor=None),
        ),
    ]
    calls: Dict[str, Any] = {"i": 0}

    def _list_backend_classes(limit: int = 50, cursor: Optional[str] = None):
        i = calls["i"]
        calls["i"] = i + 1
        p = pages[i]
        return PaginatedBackendClasses(backend_classes=p.backend_classes, pagination=p.pagination)

    monkeypatch.setattr(svc.management, "list_backend_classes", _list_backend_classes)

    out = svc.backends(online=True, device_type="QPU")
    ids = {b["id"] for b in out}
    assert ids == {"bk-1", "bk-4"}


def test_service_account_save_load(tmp_path):
    path = tmp_path / "accounts.json"
    OpenQuantumService.save_account(
        name="dev",
        token="tok_abcdefgh123456",
        filename=str(path),
        scheduler_url="https://scheduler.example.com",
        management_url="https://management.example.com",
    )
    saved = OpenQuantumService.saved_accounts(filename=str(path))
    assert "dev" in saved
    assert saved["dev"]["token"].endswith("…3456")
    svc = OpenQuantumService.from_saved_account(name="dev", filename=str(path))
    assert svc.scheduler.base_url == "https://scheduler.example.com"
    assert svc.management.base_url == "https://management.example.com"



from unittest.mock import MagicMock
from qiskit import QuantumCircuit
from qiskit.circuit import Parameter
from openquantum_sdk_qiskit import SamplerV2


def test_sampler_parameter_binding():
    backend = MagicMock()
    scheduler = MagicMock()
    scheduler.download_job_output.return_value = {"0": 100}
    mock_job = MagicMock()
    scheduler.submit_job.return_value = mock_job

    config = {
        "organization_id": "org-1",
        "backend_class_id": "backend-1",
        "job_subcategory_id": "phys:hds"
    }

    sampler = SamplerV2(backend=backend, scheduler=scheduler, config=config)

    theta = Parameter('theta')
    qc = QuantumCircuit(1)
    qc.rx(theta, 0)
    qc.measure_all()

    # PUB with parameter value
    pub = (qc, [3.14], 100)

    job = sampler.run([pub])
    job.result()

    assert scheduler.submit_job.called

    call_args = scheduler.submit_job.call_args
    config = call_args[0][0]
    assert config.job_subcategory_id == "phys:hds"

    file_content = call_args[1]['file_content']
    qasm_str = file_content.decode('utf-8')
    assert "rx(3.14" in qasm_str


def test_sampler_no_params():
    backend = MagicMock()
    scheduler = MagicMock()
    scheduler.download_job_output.return_value = {"0": 100}
    scheduler.submit_job.return_value = MagicMock()

    config = {
        "organization_id": "org-1",
        "backend_class_id": "backend-1",
        "job_subcategory_id": "phys:hds"
    }

    sampler = SamplerV2(backend=backend, scheduler=scheduler, config=config)

    qc = QuantumCircuit(1)
    qc.x(0)
    qc.measure_all()

    pub = (qc, None, 100)
    job = sampler.run([pub])
    job.result()

    assert scheduler.submit_job.called
    call_args = scheduler.submit_job.call_args
    config = call_args[0][0]
    assert config.job_subcategory_id == "phys:hds"

    file_content = call_args[1]['file_content']
    qasm_str = file_content.decode('utf-8')
    assert "x q[0];" in qasm_str or "x q0;" in qasm_str



from __future__ import annotations

from typing import Iterable, Optional, List, Dict, Any

from qiskit.primitives import (
    BaseSamplerV2,
    SamplerPubLike,
    PrimitiveResult,
    SamplerPubResult,
)
from qiskit.primitives.primitive_job import PrimitiveJob
from qiskit.providers import BackendV2
from qiskit.primitives.containers import DataBin, BitArray
from qiskit.primitives.containers.sampler_pub import SamplerPub
from ._qasm import export_circuit_to_qasm

from openquantum_sdk.clients import SchedulerClient, JobSubmissionConfig


class OQSampler(BaseSamplerV2):
    """OpenQuantum Sampler.

    This implementation submits pubs via the platform's scheduler and
    returns results using Qiskit's primitives containers (DataBin,
    SamplerPubResult, PrimitiveResult).
    """

    def __init__(
        self,
        *,
        backend: BackendV2,
        options: Optional[dict] = None,
        scheduler: Optional[SchedulerClient] = None,
        config: Optional[Dict[str, Any]] = None,
        export_format: str = "qasm3",
    ):
        self._backend = backend
        self._options = options or {}
        self._scheduler = scheduler
        self._config = config or {}
        fmt = (export_format or "qasm3").lower()
        if fmt not in {"qasm2", "qasm3"}:
            raise ValueError("export_format must be 'qasm2' or 'qasm3'")
        self._export_format = fmt

    @property
    def backend(self) -> BackendV2:
        return self._backend

    @property
    def options(self):
        return self._options

    def run(self,
            pubs: Iterable[SamplerPubLike],
            *,
            shots: int | None = None,
            export_format: Optional[str] = None,
            ) -> PrimitiveJob[PrimitiveResult[SamplerPubResult]]:
        """Execute sampler pubs and return a Qiskit PrimitiveJob."""
        if shots is None:
            shots = 100

        coerced_pubs = [SamplerPub.coerce(pub, shots) for pub in pubs]

        if not self._scheduler:
            raise RuntimeError(
                "Scheduler submission is required: provide scheduler and config when constructing OQSampler."
            )

        required = ("backend_class_id", "job_subcategory_id")
        missing = [k for k in required if not self._config.get(k)]
        if missing:
            raise ValueError(f"Missing config keys: {', '.join(missing)}")

        fmt = (export_format or self._export_format).lower()
        if fmt not in {"qasm2", "qasm3"}:
            raise ValueError("export_format must be 'qasm2' or 'qasm3'")

        def _job_fn(bound_pubs: List[SamplerPub]) -> PrimitiveResult[SamplerPubResult]:
            return self._run_via_scheduler(bound_pubs, export_fmt=fmt)

        job: PrimitiveJob[PrimitiveResult[SamplerPubResult]]
        job = PrimitiveJob(_job_fn, coerced_pubs)
        job.metadata["export_format"] = fmt
        job._submit()
        return job

    def _run_via_scheduler(self, pubs: List[SamplerPub], *, export_fmt: str) -> PrimitiveResult[SamplerPubResult]:
        """Submit pubs via scheduler and adapt results to Qiskit containers."""
        sub_id = str(self._config["job_subcategory_id"])

        pub_results: List[SamplerPubResult] = []
        for idx, pub in enumerate(pubs):
            circuit = pub.circuit
            pub_shots = pub.shots
            all_samples: List[str] = []

            flat_bindings = pub.parameter_values.ravel()

            for i in range(flat_bindings.size):
                binds = {}
                for params_tuple, val_array in flat_bindings.data.items():
                    values = val_array[i]
                    for param_name, param_val in zip(params_tuple, values):
                        binds[param_name] = param_val

                bound_circuit = circuit
                if binds:
                    bound_circuit = circuit.assign_parameters(binds)

                qasm_bytes = export_circuit_to_qasm(bound_circuit, export_fmt)

                job_config = JobSubmissionConfig(
                    organization_id=self._config.get("organization_id"),
                    backend_class_id=self._config["backend_class_id"],
                    name=self._config.get("name", f"OQ Sampler Pub {idx} Sample {i}"),
                    job_subcategory_id=sub_id,
                    shots=pub_shots,
                    configuration_data=self._config.get("configuration_data", {}),
                )

                job = self._scheduler.submit_job(job_config, file_content=qasm_bytes)
                counts: Dict[str, int] = self._scheduler.download_job_output(job)
                samples = self._counts_to_samples(counts)
                if len(samples) != pub_shots:
                    if len(samples) < pub_shots:
                        samples.extend(["0" * bound_circuit.num_clbits] * (pub_shots - len(samples)))
                    else:
                        samples = samples[:pub_shots]

                all_samples.extend(samples)

            num_bits = circuit.num_clbits if circuit.num_clbits > 0 else max(len(s) for s in all_samples)
            ba = BitArray.from_samples(all_samples, num_bits)
            expected_shape = pub.parameter_values.shape + (pub_shots,)
            ba = ba.reshape(expected_shape)

            meas_container = DataBin(meas=ba, shape=pub.parameter_values.shape)
            pub_results.append(
                SamplerPubResult(
                    meas_container,
                    metadata={
                        "shots": pub_shots,
                        "circuit_metadata": getattr(circuit, "metadata", None),
                    },
                )
            )

        return PrimitiveResult(pub_results, metadata={"version": 2})

    def _counts_to_samples(self, counts: Dict[str, int]) -> List[str]:
        samples: List[str] = []
        for bitstr, c in counts.items():
            samples.extend([bitstr] * int(c))
        return samples



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional

from openquantum_sdk.qiskit import OpenQuantumService
from openquantum_sdk.models import (
    BackendClassRead,
    PaginatedBackendClasses,
    PaginationInfo,
)


@dataclass
class _Page:
    backend_classes: List[BackendClassRead]
    pagination: PaginationInfo


def _bk(idx: int, **kw) -> BackendClassRead:
    return BackendClassRead(
        id=f"bk-{idx}",
        name=kw.get("name", f"Backend {idx}"),
        description="",
        type=kw.get("type", "QPU"),
        provider_id=kw.get("provider_id", "prov-1"),
        short_code=kw.get("short_code", f"sc-{idx}"),
        queue_depth=kw.get("queue_depth", 0),
        accepting_jobs=kw.get("accepting_jobs", True),
        status=kw.get("status", "Online"),
    )


def test_service_backends_all_filters(monkeypatch):
    svc = OpenQuantumService()

    pages = [
        _Page(
            backend_classes=[
                _bk(1, name="Alpha", short_code="ALP", provider_id="prov-A", min_qubits=4),
                _bk(2, name="Bravo", type="SIMULATOR", provider_id="prov-B", min_qubits=8),
            ],
            pagination=PaginationInfo(next_cursor="cursor-2"),
        ),
        _Page(
            backend_classes=[
                _bk(3, name="Charlie", provider_id="prov-A", accepting_jobs=False, status="Offline", min_qubits=16),
                _bk(4, name="Delta", provider_id="prov-A", min_qubits=32),
            ],
            pagination=PaginationInfo(next_cursor=None),
        ),
    ]

    def _list_backend_classes(limit: int = 50, cursor: Optional[str] = None):
        if not cursor:
            p = pages[0]
        else:
            p = pages[1]
        return PaginatedBackendClasses(backend_classes=p.backend_classes, pagination=p.pagination)

    monkeypatch.setattr(svc.management, "list_backend_classes", _list_backend_classes)

    out_name = svc.backends(name="alp")
    assert {b["id"] for b in out_name} == {"bk-1"}

    out_online_qpu = svc.backends(online=True, device_type="QPU")
    assert {b["id"] for b in out_online_qpu} == {"bk-1", "bk-4"}

    out_vendor = svc.backends(device_type="QPU", vendor_id="prov-A")
    assert {b["id"] for b in out_vendor} == {"bk-1", "bk-4", "bk-3"}



import pytest
from openquantum_sdk_qiskit.oq_service import OpenQuantumService
from openquantum_sdk_qiskit.oq_backend import OpenQuantumBackend


def test_service_return_backend_hardcoded():
    with pytest.MonkeyPatch.context() as m:
        m.setenv("OPENQUANTUM_NO_AUTOLOAD", "1")
        service = OpenQuantumService(token="dummy")

        backend = service.return_backend("Rigetti Ankaa-3")

        assert isinstance(backend, OpenQuantumBackend)
        assert backend.name == "Rigetti Ankaa-3"
        assert backend.num_qubits == 84
        assert "rx" in backend.target.operation_names


def test_service_return_backend_ionq():
    with pytest.MonkeyPatch.context() as m:
        m.setenv("OPENQUANTUM_NO_AUTOLOAD", "1")
        service = OpenQuantumService(token="dummy")

        backend = service.return_backend("IonQ Aria-1")
        assert backend.num_qubits == 25
        assert "ms" in backend.target.operation_names



from __future__ import annotations

from .oq_service import OpenQuantumService
from .oq_backend import OpenQuantumBackend
from .oq_sampler import OQSampler
from .oq_estimator import OQEstimator

SamplerV2 = OQSampler
EstimatorV2 = OQEstimator

__version__ = "0.2.4"
__all__ = [
    "OpenQuantumService",
    "OpenQuantumBackend",
    "OQSampler",
    "SamplerV2",
    "OQEstimator",
    "EstimatorV2",
]



from __future__ import annotations

import datetime
from typing import Any, Dict, Iterable, List, Optional

from qiskit.providers import BackendV2, Options
from qiskit.result import Result

from .oq_target import build_target_from_capabilities
from ._qasm import export_circuit_to_qasm
from openquantum_sdk.clients import SchedulerClient, JobSubmissionConfig
from .backend_data import (
    RIGETTI_ANKAA_3_CAPS,
    IONQ_ARIA_1_CAPS,
    IONQ_FORTE_1_CAPS,
    IQM_EMERALD_CAPS,
    IQM_GARNET_CAPS,
    AQT_IBEX_Q1_CAPS
)


class OpenQuantumBackend(BackendV2):
    """Qiskit Backend for the OpenQuantum platform.
    """

    def __init__(
        self,
        name: str,
        *,
        capabilities: Optional[Dict[str, Any]] = None,
        scheduler: Optional[SchedulerClient] = None,
        config: Optional[Dict[str, Any]] = None,
        export_format: str = "qasm3",
        **kwargs
    ) -> None:
        super().__init__(name=name, **kwargs)

        self._scheduler = scheduler
        self._config = config or {}
        fmt = (export_format or "qasm3").lower()
        if fmt not in {"qasm2", "qasm3"}:
            raise ValueError("export_format must be 'qasm2' or 'qasm3'")
        self._export_format = fmt

        if capabilities is None:
            name_lower = name.lower()
            if "ankaa" in name_lower:
                capabilities = RIGETTI_ANKAA_3_CAPS
            elif "aria" in name_lower:
                capabilities = IONQ_ARIA_1_CAPS
            elif "forte" in name_lower:           # Going to need to be careful with this logic should we ever add Forte Enterprise
                capabilities = IONQ_FORTE_1_CAPS
            elif "emerald" in name_lower:
                capabilities = IQM_EMERALD_CAPS
            elif "garnet" in name_lower:
                capabilities = IQM_GARNET_CAPS
            elif "ibex_q1" in name_lower:
                capabilities = AQT_IBEX_Q1_CAPS
            else:
                capabilities = {
                    "n_qubits": 1,
                    "native_ops": [{"name": "measure", "arity": 1}],
                    "topology": {"directed_edges": False, "coupling_map": []}
                }

        self._cap = capabilities
        self._target = build_target_from_capabilities(capabilities)
        self._max_circuits = (self._cap.get("limits", {}) or {}).get("max_circuits", 1024)

        self.description = f"OpenQuantum Backend: {self.name}"
        self.online_date = datetime.datetime.now()
        self.backend_version = "0.0.1"

    @property
    def target(self):
        return self._target

    @property
    def max_circuits(self):
        return self._max_circuits

    @classmethod
    def _default_options(cls):
        return Options(shots=1024)

    def run(self, circuits: Iterable[Any], /, **run_options):
        """Submit circuits to platform and return a Qiskit Job-like object.

        BackendSamplerV2 compatibility - yields a Result with
        experiment data.memory as list[str] bitstrings.
        """
        if not self._scheduler:
            raise RuntimeError(
                "Scheduler submission is required: provide scheduler and config when constructing backend."
            )

        required = ("backend_class_id", "job_subcategory_id")
        missing = [k for k in required if not self._config.get(k)]
        if missing:
            raise ValueError(f"Missing config keys: {', '.join(missing)}")

        shots = run_options.get("shots") or getattr(self.options, "shots", None) or 100
        memory = run_options.get("memory", True)
        export_format = (run_options.get("export_format") or self._export_format).lower()

        if not isinstance(circuits, list):
            circuits = [circuits]

        results_payload: List[Dict[str, Any]] = []
        sub_id = str(self._config["job_subcategory_id"])

        for idx, circ in enumerate(circuits):
            payload = export_circuit_to_qasm(circ, export_format)

            job_config = JobSubmissionConfig(
                organization_id=self._config.get("organization_id"),
                backend_class_id=self._config["backend_class_id"],
                name=self._config.get("name", f"OQ Backend Run {idx}"),
                job_subcategory_id=sub_id,
                shots=shots,
                configuration_data=self._config.get("configuration_data", {}),
            )
            job = self._scheduler.submit_job(job_config, file_content=payload)
            counts: Dict[str, int] = self._scheduler.download_job_output(job)

            bit_width = max((len(k) for k in counts.keys()), default=0)
            memory_hex = self._counts_to_memory_hex(counts, bit_width)
            counts_hex = self._counts_bits_to_hex_keys(counts)

            results_payload.append(
                {
                    "shots": shots,
                    "success": True,
                    "data": {
                        "memory": memory_hex if memory and memory_hex is not None else None,
                        "counts": counts_hex,
                    },
                }
            )

        return _ImmediateResultJob(
            Result.from_dict(
                {
                    "results": results_payload,
                    "backend_name": self.name,
                    "success": True,
                    "backend_version": "0.1.0",
                    "job_id": "oq_job_id_placeholder",
                }
            )
        )

    # TODO: Job batching support

    def _counts_to_memory_hex(self, counts: Dict[str, int], bit_width: int) -> List[str]:
        """Expand counts of bitstrings into a list of hex memory strings."""
        memory: List[str] = []
        for bitstr, c in counts.items():
            s = bitstr.strip()
            if bit_width and len(s) < bit_width:
                s = s.zfill(bit_width)
            try:
                h = hex(int(s, 2))
                memory.extend([h] * int(c))
            except ValueError:
                pass
        return memory

    def _counts_bits_to_hex_keys(self, counts: Dict[str, int]) -> Dict[str, int]:
        out: Dict[str, int] = {}
        for bitstr, c in counts.items():
            try:
                h = hex(int(bitstr, 2))
                out[h] = int(c)
            except ValueError:
                pass
        return out


class _ImmediateResultJob:
    def __init__(self, result: Result):
        self._result = result

    def result(self) -> Result:
        return self._result

    def status(self):
        from qiskit.providers import JobStatus
        return JobStatus.DONE

    def job_id(self):
        return self._result.job_id



from typing import Optional, Dict, Any, Iterable
import copy

from qiskit.primitives import BackendEstimatorV2, BaseEstimatorV2, PrimitiveResult, PubResult, EstimatorPubLike
from qiskit.primitives.primitive_job import PrimitiveJob
from qiskit.providers import BackendV2
from openquantum_sdk.clients import SchedulerClient


class OQEstimator(BaseEstimatorV2):
    """OpenQuantum Estimator.

    This implementation uses Qiskit's BackendEstimatorV2 logic to compute
    expectation values from shots execution on the OpenQuantum backend.
    """

    def __init__(
        self,
        *,
        backend: BackendV2,
        options: Optional[dict] = None,
        scheduler: Optional[SchedulerClient] = None,
        config: Optional[Dict[str, Any]] = None,
        export_format: str = "qasm3",
    ):
        """Initialize the Estimator.

        Args:
            backend: The backend to run circuits on.
            options: Primitive options.
            scheduler: (Optional) Explicit scheduler client (usually passed to backend).
            config: (Optional) Explicit config (usually passed to backend).
            export_format: (Optional) Export format.
        """
        self._backend = backend
        self._scheduler = scheduler
        self._config = config or {}
        self._export_format = export_format
        self._default_precision = 0.01
        backend_proxy = copy.copy(backend)

        if self._config and hasattr(backend_proxy, "_config"):
            current = getattr(backend_proxy, "_config", {}) or {}
            merged = {**current, **self._config}
            backend_proxy._config = merged

        if self._scheduler and hasattr(backend_proxy, "_scheduler"):
            backend_proxy._scheduler = self._scheduler

        self._backend_estimator = BackendEstimatorV2(backend=backend_proxy, options=options)

    @property
    def backend(self) -> BackendV2:
        return self._backend

    @property
    def options(self):
        """Return the options for the estimator."""
        return self._backend_estimator.options

    def run(
        self, pubs: Iterable[EstimatorPubLike], *, precision: float | None = None
    ) -> PrimitiveJob[PrimitiveResult[PubResult]]:
        return self._backend_estimator.run(pubs, precision=precision)



from __future__ import annotations

from typing import Any, Dict  # noqa: F401

from openquantum_sdk.qiskit import OpenQuantumService
from openquantum_sdk.auth import ClientCredentials


def test_service_login_token_saves_and_active_account(tmp_path):
    path = tmp_path / "accounts.json"
    svc = OpenQuantumService.login(
        token="tok_abcdefgh123456",
        save=True,
        name="dev",
        filename=str(path),
        scheduler_url="https://scheduler.example.com",
        management_url="https://management.example.com",
    )
    assert svc.active_account["auth_mode"] == "token"

    saved = OpenQuantumService.saved_accounts(filename=str(path))
    assert "dev" in saved
    assert saved["dev"]["token"].endswith("…3456")


def test_client_credentials_auth_applies_header_and_caches(monkeypatch):
    calls: Dict[str, Any] = {"n": 0}

    class _Resp:
        status_code = 200

        def json(self):
            return {"access_token": "abc123", "expires_in": 300}

        def raise_for_status(self):
            return None

    def _fake_post(self, url, data=None, timeout=None, **kwargs):
        calls["n"] += 1
        return _Resp()

    import requests

    monkeypatch.setattr(requests.Session, "post", _fake_post, raising=False)

    svc = OpenQuantumService(
        creds=ClientCredentials(client_id="cid", client_secret="secret"),
        keycloak_base="https://id.example.com",
        realm="platform",
    )

    headers1 = svc.management._authorized_headers({})
    assert headers1.get("Authorization") == "Bearer abc123"
    assert calls["n"] == 1

    headers2 = svc.management._authorized_headers({})
    assert headers2.get("Authorization") == "Bearer abc123"
    assert calls["n"] == 1



from __future__ import annotations

from typing import Any, Dict, List, Optional, Union, TYPE_CHECKING
from pathlib import Path
import json
import os

from openquantum_sdk.auth import ClientCredentials, ClientCredentialsAuth
from openquantum_sdk.clients import ManagementClient, SchedulerClient

if TYPE_CHECKING:
    from qiskit.providers import BackendV2


def _require_qiskit() -> None:
    try:
        import qiskit  # noqa: F401
    except Exception as e:
        raise RuntimeError(
            "Qiskit is required for this operation. Install: pip install 'openquantum-sdk[qiskit]'"
        ) from e


class OpenQuantumService:
    """service for auth and backend availability discovery"""

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        creds: Optional[ClientCredentials] = None,
        keycloak_base: str = "https://id.openquantum.com",
        realm: str = "platform",
        scheduler_url: str = "https://scheduler.openquantum.com",
        management_url: str = "https://management.openquantum.com",
    ) -> None:

        auth_obj: Optional[ClientCredentialsAuth] = None
        if token is None and creds is None and os.getenv("OPENQUANTUM_NO_AUTOLOAD") != "1":
            try:
                acct_name = os.getenv("OPENQUANTUM_ACCOUNT_NAME", "default")
                path = self._accounts_path(filename=None)
                if path.exists():
                    data = json.loads(path.read_text(encoding="utf-8"))
                    if acct_name in data:
                        entry = data[acct_name]
                        kr_used = bool(entry.get("keyring", False))
                        if (entry.get("auth_mode") == "token"):
                            token = entry.get("token")
                            if (not token) and kr_used:
                                try:
                                    import keyring
                                    token = keyring.get_password("openquantum", f"{acct_name}:token")
                                except Exception:
                                    pass
                        else:
                            c = (entry.get("creds") or {})
                            if c.get("client_id"):
                                secret = c.get("client_secret") or ""
                                if (not secret) and kr_used:
                                    try:
                                        import keyring
                                        maybe = keyring.get_password("openquantum", f"{acct_name}:client_secret")
                                        secret = maybe or secret
                                    except Exception:
                                        pass
                                creds = ClientCredentials(client_id=c.get("client_id"), client_secret=secret)
                        keycloak_base = entry.get("keycloak_base", keycloak_base)
                        realm = entry.get("realm", realm)
                        scheduler_url = entry.get("scheduler_url", scheduler_url)
                        management_url = entry.get("management_url", management_url)
            except Exception as e:
                import warnings as _warnings
                _warnings.warn(
                    f"OpenQuantumService autoload failed: {e!r}. Continuing unauthenticated.",
                    stacklevel=2,
                )
        if token and creds:
            raise ValueError("Provide either 'token' or 'creds', not both.")
        if creds is not None:
            auth_obj = ClientCredentialsAuth(
                creds=creds,
                keycloak_base=keycloak_base,
                realm=realm,
            )

        self.management = ManagementClient(base_url=management_url, token=token, auth=auth_obj)
        self.scheduler = SchedulerClient(base_url=scheduler_url, token=token, auth=auth_obj)

    def close(self) -> None:
        """Close underlying HTTP sessions."""
        self.management.close()
        self.scheduler.close()

    def backends(
        self,
        *,
        name: Optional[str] = None,
        online: Optional[bool] = None,
        device_type: Optional[str] = None,  # "QPU" or "SIMULATOR"
        vendor_id: Optional[str] = None,
        min_num_qubits: Optional[int] = None,
        limit: int = 50,
    ) -> List[Dict[str, Any]]:
        """List backend classes with availability info.

        Filters:
            - name: substring match against backend class name
            - online: if True, require status == "Online" and accepting_jobs == True
            - device_type: match ``type`` field from Management API
            - vendor_id: match ``provider_id``
            - min_num_qubits: DEPRECATED (ignored) - constraint_data removed in SDK v0.2.1
        """
        items: List[Dict[str, Any]] = []
        cursor: Optional[str] = None
        while True:
            page = self.management.list_backend_classes(limit=limit, cursor=cursor)
            for bc in page.backend_classes:
                if name:
                    needle = name.lower()
                    hay_name = (bc.name or "").lower()
                    hay_short = (bc.short_code or "").lower()
                    if needle not in hay_name and needle not in hay_short:
                        continue
                if device_type and (bc.type or "").lower() != device_type.lower():
                    continue
                if vendor_id and (bc.provider_id or "") != vendor_id:
                    continue
                if online is True and (((bc.status or "").lower() != "online") or not bool(bc.accepting_jobs)):
                    continue
                # Note: min_num_qubits filter removed as constraint_data no longer exists in SDK v0.2.1

                items.append(
                    {
                        "id": bc.id,
                        "name": bc.name,
                        "description": bc.description,
                        "type": bc.type,
                        "provider_id": bc.provider_id,
                        "short_code": bc.short_code,
                        "queue_depth": bc.queue_depth,
                        "accepting_jobs": bc.accepting_jobs,
                        "status": bc.status,
                    }
                )
            cursor = page.pagination.next_cursor
            if not cursor:
                break
        return items

    def return_target(
        self,
        name: str,
        *,
        capabilities_source: Optional[Union[str, Path, Dict[str, Any]]] = None,
    ) -> Any:  # target primative

        _require_qiskit()
        from .oq_target import build_target_from_capabilities
        caps = self._load_capabilities(name, capabilities_source)

        return build_target_from_capabilities(caps)

    def return_backend(
        self,
        name: str,
        *,
        capabilities_source: Optional[Union[str, Path, Dict[str, Any]]] = None,
        config: Optional[Dict[str, Any]] = None,
        export_format: str = "qasm3",
    ) -> Any:  # BackendV2 object
        """
        Args:
            name: backend identifier (id/short_code/name) or label.
            capabilities_source: dict|path|URL to capabilities; if None, resolves via Management API.
            config: required IDs for platform submission (org, backend_class_id, subcategory).
            export_format: 'qasm2' or 'qasm3'.
        """
        _require_qiskit()
        from .oq_backend import OpenQuantumBackend

        caps = self._load_capabilities(name, capabilities_source)
        backend = OpenQuantumBackend(
            name=name,
            capabilities=caps,
            scheduler=self.scheduler,
            config=config,
            export_format=export_format,
        )
        return backend

    def create_sampler(
        self,
        *,
        backend: "BackendV2",
        organization_id: Optional[str] = None,
        backend_class_id: str,
        job_subcategory_id: str,
        name: Optional[str] = None,
        configuration_data: Optional[Dict[str, Any]] = None,
        export_format: str = "qasm3",
    ):
        """Create a SamplerV2 (OQSampler internally)

        Args:
            backend: A Qiskit BackendV2 to transpile/run against.
            organization_id: Optional organization UUID.
            backend_class_id: Target backend class (UUID or short code).
            job_subcategory_id: Workload subcategory (UUID or short code).
            name: Optional job name prefix.
            configuration_data: Optional provider-specific configuration.
            export_format: "qasm2" or "qasm3" for submission IR.

        Returns:
            An initialized SamplerV2 that submits jobs to OpenQuantum.
        """
        _require_qiskit()
        try:
            from .oq_sampler import OQSampler
        except ImportError as e:
            raise ImportError("Qiskit plugin not installed correctly") from e

        config = {
            "organization_id": organization_id,
            "backend_class_id": backend_class_id,
            "job_subcategory_id": job_subcategory_id,
            "name": name,
            "configuration_data": configuration_data,
        }

        return OQSampler(
            backend=backend,
            scheduler=self.scheduler,
            config=config,
            export_format=export_format,
        )

    def create_estimator(
        self,
        *,
        backend: "BackendV2",
        organization_id: Optional[str] = None,
        backend_class_id: str,
        job_subcategory_id: str,
        name: Optional[str] = None,
        configuration_data: Optional[Dict[str, Any]] = None,
        export_format: str = "qasm3",
    ):
        """Create an EstimatorV2 (OQEstimator internally)

        Args:
            backend: A Qiskit BackendV2 to transpile/run against.
            organization_id: Optional organization UUID.
            backend_class_id: Target backend class (UUID or short code).
            job_subcategory_id: Workload subcategory (UUID or short code).
            name: Optional job name prefix.
            configuration_data: Optional provider-specific configuration.
            export_format: "qasm3" (default) or "qasm2".

        Returns:
            An initialized EstimatorV2 that submits jobs to OpenQuantum.
        """
        _require_qiskit()
        try:
            from .oq_estimator import OQEstimator
        except ImportError as e:
            raise ImportError("Qiskit plugin not installed correctly") from e

        config = {
            "organization_id": organization_id,
            "backend_class_id": backend_class_id,
            "job_subcategory_id": job_subcategory_id,
            "name": name,
            "configuration_data": configuration_data,
        }

        return OQEstimator(
            backend=backend,
            scheduler=self.scheduler,
            config=config,
            export_format=export_format,
        )

    def _load_capabilities(
        self,
        name: str,
        src: Optional[Union[str, Path, Dict[str, Any]]],
    ) -> Dict[str, Any]:
        """Capability loader (placeholder).
           TODO: Wire to a real capabilities endpoint.
        """
        from .backend_data import (
            RIGETTI_ANKAA_3_CAPS,
            IONQ_ARIA_1_CAPS,
            IONQ_FORTE_1_CAPS,
            IQM_EMERALD_CAPS,
            IQM_GARNET_CAPS,
            AQT_IBEX_Q1_CAPS
        )

        n = name.lower()
        if "ankaa-3" in n or "rigetti" in n:
            return RIGETTI_ANKAA_3_CAPS
        if "aria-1" in n or "aria" in n:
            return IONQ_ARIA_1_CAPS
        if "forte" in n:              # Double check if Forte enterprise has the same caps when/if we add it
            return IONQ_FORTE_1_CAPS
        if "emerald" in n:
            return IQM_EMERALD_CAPS
        if "garnet" in n:
            return IQM_GARNET_CAPS
        if "ibex-q1" in n or "ibex" in n:
            return AQT_IBEX_Q1_CAPS

        if isinstance(src, dict):
            return src

        if isinstance(src, (str, Path)):
            p = Path(src)
            if p.exists():
                return json.loads(p.read_text(encoding="utf-8"))

            import requests
            r = requests.get(str(src), timeout=15)
            r.raise_for_status()

            return r.json()

        page = self.management.list_backend_classes(limit=100)
        match = next(
            (
                bc
                for bc in page.backend_classes
                if bc.id == name or (bc.short_code or "") == name or (bc.name or "") == name
            ),
            None,
        )
        # TODO: replace with real field
        if not match or not getattr(match, "extra", {}).get("capabilities_url"):
            raise ValueError(f"Capabilities for backend '{name}' not found")

        import requests
        url = match.extra["capabilities_url"]
        r = requests.get(url, timeout=15)
        r.raise_for_status()

        return r.json()

    @property
    def active_account(self) -> Dict[str, Any]:
        """Return a minimal description of the active auth context."""

        auth_mode = "unauthenticated"
        if getattr(self.management, "_fixed_token", None):
            auth_mode = "token"
        elif getattr(self.management, "_auth", None):
            auth_mode = "client_credentials"
        return {
            "management_url": self.management.base_url,
            "scheduler_url": self.scheduler.base_url,
            "auth_mode": auth_mode,
        }

    @staticmethod
    def _accounts_path(filename: Optional[str]) -> Path:
        if filename:
            return Path(filename).expanduser()
        cfg_dir = os.getenv("OPENQUANTUM_CONFIG_DIR")
        if cfg_dir:
            return Path(cfg_dir).expanduser() / "accounts.json"
        try:
            from platformdirs import PlatformDirs

            d = PlatformDirs(appname="OpenQuantum", appauthor=False)
            return Path(d.user_config_dir) / "accounts.json"
        except Exception:
            return Path(os.path.expanduser("~/.openquantum/accounts.json"))

    @classmethod
    def save_account(
        cls,
        *,
        name: str = "default",
        token: Optional[str] = None,
        creds: Optional[ClientCredentials] = None,
        keycloak_base: str = "https://id.openquantum.com",
        realm: str = "platform",
        scheduler_url: str = "https://scheduler.openquantum.com",
        management_url: str = "https://management.openquantum.com",
        filename: Optional[str] = None,
        overwrite: bool = True,
        use_keyring: bool = True,
    ) -> None:
        """alt to qiskit account manager

        for local dev
        """
        if token and creds:
            raise ValueError("Provide either 'token' or 'creds', not both.")
        path = cls._accounts_path(filename)
        path.parent.mkdir(parents=True, exist_ok=True)
        data: Dict[str, Any] = {}
        if path.exists():
            try:
                data = json.loads(path.read_text(encoding="utf-8"))
            except Exception:
                if not overwrite:
                    raise
                data = {}
        if not overwrite and name in data:
            raise ValueError(f"Account '{name}' already exists. Set overwrite=True to replace.")
        entry: Dict[str, Any] = {
            "auth_mode": "token" if token else "client_credentials",
            "token": token,
            "creds": {"client_id": creds.client_id, "client_secret": creds.client_secret} if creds else None,
            "keycloak_base": keycloak_base,
            "realm": realm,
            "scheduler_url": scheduler_url,
            "management_url": management_url,
        }
        if use_keyring:
            try:
                import keyring

                if token:
                    keyring.set_password("openquantum", f"{name}:token", token)
                    entry["token"] = None
                if creds and creds.client_secret:
                    keyring.set_password("openquantum", f"{name}:client_secret", creds.client_secret)
                    entry["creds"] = {"client_id": creds.client_id, "client_secret": None}
                entry["keyring"] = True
            except Exception:
                entry.setdefault("keyring", False)
        data[name] = entry

        tmp_txt = json.dumps(data, indent=2)
        from tempfile import NamedTemporaryFile as _NTF

        cls._ensure_cfg_dir(path)
        with _NTF("w", delete=False, encoding="utf-8", dir=str(path.parent)) as _tmp:
            _tmp.write(tmp_txt)
            _tmp.flush()

            os.fsync(_tmp.fileno())
            _tmp_path = Path(_tmp.name)
        _tmp_path.replace(path)
        try:
            os.chmod(path, 0o600)
        except Exception:
            pass

    @staticmethod
    def _ensure_cfg_dir(path: Path) -> None:
        try:
            path.parent.mkdir(parents=True, exist_ok=True)
        except Exception:
            pass

    @classmethod
    def saved_accounts(
        cls,
        *,
        filename: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Return all saved accounts"""
        path = cls._accounts_path(filename)
        if not path.exists():
            return {}
        try:
            raw = json.loads(path.read_text(encoding="utf-8"))
            masked: Dict[str, Any] = {}

            for k, v in raw.items():
                v = dict(v)
                tok = v.get("token")

                if tok:
                    v["token"] = tok[:6] + "…" + tok[-4:]

                creds = v.get("creds") or {}
                if creds.get("client_secret"):
                    cs = creds["client_secret"]
                    creds["client_secret"] = cs[:4] + "…" + cs[-4:]

                v["creds"] = creds
                masked[k] = v
            return masked

        except Exception:
            return {}

    @classmethod
    def delete_account(
        cls,
        *,
        name: str = "default",
        filename: Optional[str] = None,
    ) -> bool:
        """Delete a saved account by name."""
        path = cls._accounts_path(filename)
        if not path.exists():
            return False
        try:
            data = json.loads(path.read_text(encoding="utf-8"))
        except Exception:
            return False
        if name not in data:
            return False
        try:
            import keyring

            keyring.delete_password("openquantum", f"{name}:token")
            keyring.delete_password("openquantum", f"{name}:client_secret")
        except Exception:
            pass
        del data[name]
        path.write_text(json.dumps(data, indent=2), encoding="utf-8")
        return True

    @classmethod
    def from_saved_account(
        cls,
        *,
        name: str = "default",
        filename: Optional[str] = None,
    ) -> "OpenQuantumService":
        """Instantiate service from a saved account configuration."""
        path = cls._accounts_path(filename)
        if not path.exists():
            raise ValueError(f"Saved account file not found: {path}")
        try:
            data = json.loads(path.read_text(encoding="utf-8"))
        except Exception as e:
            raise ValueError(f"Failed to load saved accounts from {path}: {e}") from e
        if name not in data:
            raise ValueError(f"Saved account '{name}' not found")
        entry = data[name]
        kr = bool(entry.get("keyring", False))
        if entry.get("auth_mode") == "token":
            tok = entry.get("token")
            if (not tok) and kr:
                try:
                    import keyring

                    tok = keyring.get_password("openquantum", f"{name}:token")
                except Exception:
                    pass
            return cls(
                token=tok,
                scheduler_url=entry.get("scheduler_url"),
                management_url=entry.get("management_url"),
            )
        creds_dict = entry.get("creds") or {}
        client_secret = creds_dict.get("client_secret", "")
        if (not client_secret) and kr:
            try:
                import keyring

                maybe = keyring.get_password("openquantum", f"{name}:client_secret")
                client_secret = maybe or client_secret
            except Exception:
                pass
        creds = ClientCredentials(
            client_id=creds_dict.get("client_id", ""),
            client_secret=client_secret,
        )
        return cls(
            creds=creds,
            keycloak_base=entry.get("keycloak_base", "https://id.openquantum.com"),
            realm=entry.get("realm", "platform"),
            scheduler_url=entry.get("scheduler_url"),
            management_url=entry.get("management_url"),
        )

    @classmethod
    def login(
        cls,
        *,
        token: Optional[str] = None,
        creds: Optional[ClientCredentials] = None,
        save: bool = False,
        name: str = "default",
        filename: Optional[str] = None,
        **kwargs: Any,
    ) -> "OpenQuantumService":

        svc = cls(token=token, creds=creds, **kwargs)

        if save:
            cls.save_account(
                name=name,
                token=token,
                creds=creds,
                keycloak_base=kwargs.get("keycloak_base", "https://id.openquantum.com"),
                realm=kwargs.get("realm", "platform"),
                scheduler_url=kwargs.get("scheduler_url", "https://scheduler.openquantum.com"),
                management_url=kwargs.get("management_url", "https://management.openquantum.com"),
                filename=filename,
                overwrite=True,
                use_keyring=True,
            )
        return svc

    def __enter__(self) -> "OpenQuantumService":
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()



from __future__ import annotations
from openquantum_sdk.models import JobRead
from openquantum_sdk.clients import SchedulerClient
from openquantum_sdk_qiskit.oq_backend import OpenQuantumBackend
from openquantum_sdk_qiskit.oq_sampler import OQSampler
from openquantum_sdk_qiskit.oq_target import build_target_from_capabilities
from qiskit.primitives import BackendSamplerV2
from qiskit import QuantumCircuit

import json
from typing import Dict

import pytest

qiskit = pytest.importorskip("qiskit")


def _caps_2q() -> Dict:
    return {
        "n_qubits": 2,
        "timing": {"dt": 2.0e-9, "durations": [{"op": "cx", "qubits": [0, 1], "value_s": 5.0e-7}]},
        "native_ops": [
            {"name": "x", "arity": 1},
            {"name": "h", "arity": 1},
            {"name": "cx", "arity": 2},
        ],
        "topology": {"directed_edges": False, "coupling_map": [[0, 1]]},
        "noise": {"gate_error": [{"op": "cx", "qubits": [0, 1], "prob": 0.01}]},
        "limits": {"max_circuits": 100},
    }


def test_target_builder_basic():
    cap = _caps_2q()
    target = build_target_from_capabilities(cap)
    assert target.num_qubits == 2
    # cx on (0,1) and (1,0)
    assert "cx" in target.operation_names
    assert (0, 1) in target["cx"].keys()
    assert (1, 0) in target["cx"].keys()


def test_sampler_scheduler_counts_json_to_primitive_result(requests_mock):
    qc = QuantumCircuit(2)
    qc.h(0)
    qc.cx(0, 1)
    qc.measure_all()

    output_url = "https://output.example.com/job-abc"
    counts = {"00": 50, "11": 50}
    requests_mock.get(output_url, text=json.dumps(counts))

    class _SchedulerStub(SchedulerClient):
        base_url = "https://scheduler.openquantum.com"

        def __init__(self, output_url: str, counts: Dict[str, int]):
            self._output_url = output_url
            self._counts = counts

        def submit_job(self, config, file_content=None, file_path=None):
            return JobRead(
                id="job-adapter",
                status="Completed",
                input_data_url="https://input.example.com/job-adapter",
                job_preparation_id="prep",
                execution_plan_id="plan",
                queue_priority_id="prio",
                message=None,
                output_data_url=self._output_url,
                transaction_id=None,
                submitted_at=None,
            )

        def get_job_categories(self, limit=100):
            from unittest.mock import MagicMock
            cats = MagicMock()
            c = MagicMock()
            c.id = "cat-1"
            c.name = "Qiskit"
            cats.categories = [c]
            return cats

        def get_job_subcategories(self, category_id, limit=100):
            from unittest.mock import MagicMock
            subs = MagicMock()
            s = MagicMock()
            s.id = "sub-1"
            s.name = "sub-1"
            subs.categories = [s]
            return subs

    sampler = OQSampler(
        backend=object(),
        scheduler=_SchedulerStub(output_url, counts),
        config={
            "organization_id": "org-1",
            "backend_class_id": "bk-1",
            "job_subcategory_id": "sub-1",
        },
        export_format="qasm3",
    )
    job = sampler.run([(qc, None, 100)])
    res = job.result()
    assert len(res) == 1
    pub0 = res[0]
    # DataBin should have BitArray shots
    total = sum(pub0.data.meas.get_counts().values())
    assert total == 100


def test_backend_run_parses_counts_to_memory(requests_mock):
    cap = _caps_2q()
    backend = OpenQuantumBackend(
        name="oq://test/backend",
        capabilities=cap,
        scheduler=None,
        config=None,
    )

    class _SchedulerStub(SchedulerClient):
        base_url = "https://scheduler.openquantum.com"

        def __init__(self):
            pass

        def submit_job(self, config, file_content=None, file_path=None):
            return JobRead(
                id="job-xyz",
                status="Completed",
                input_data_url="https://input.example.com/job-xyz",
                job_preparation_id="prep-xyz",
                execution_plan_id="plan",
                queue_priority_id="prio",
                message=None,
                output_data_url="https://output.example.com/job-xyz",
                transaction_id=None,
                submitted_at=None,
            )

        def get_job_categories(self, limit=100):
            from unittest.mock import MagicMock
            cats = MagicMock()
            c = MagicMock()
            c.id = "cat-1"
            c.name = "Qiskit"
            cats.categories = [c]
            return cats

        def get_job_subcategories(self, category_id, limit=100):
            from unittest.mock import MagicMock
            subs = MagicMock()
            s = MagicMock()
            s.id = "sub-1"
            s.name = "sub-1"
            subs.categories = [s]
            return subs

    backend._scheduler = _SchedulerStub()
    backend._config = {
        "organization_id": "org-1",
        "backend_class_id": "bk-1",
        "job_subcategory_id": "sub-1",
    }

    counts = {"00": 60, "11": 40}
    requests_mock.get("https://output.example.com/job-xyz", text=json.dumps(counts))

    qc = QuantumCircuit(2)
    qc.h(0)
    qc.cx(0, 1)
    qc.measure_all()

    job = backend.run([qc], shots=100, memory=True)
    result = job.result()
    assert hasattr(result, "results")
    assert len(result.results) == 1
    exp0 = result.results[0]
    assert exp0.shots == 100
    assert isinstance(exp0.data.memory, list)
    assert len(exp0.data.memory) == 100


def test_backend_sampler_v2_integration_with_backend_run(requests_mock):
    cap = _caps_2q()
    backend = OpenQuantumBackend(
        name="oq://test/backend",
        capabilities=cap,
        scheduler=None,
        config=None,
    )

    class _SchedulerStub(SchedulerClient):
        base_url = "https://scheduler.openquantum.com"

        def __init__(self):
            pass

        def submit_job(self, config, file_content=None, file_path=None):
            return JobRead(
                id="job-s2v",
                status="Completed",
                input_data_url="https://input.example.com/job-s2v",
                job_preparation_id="prep-s2v",
                execution_plan_id="plan",
                queue_priority_id="prio",
                message=None,
                output_data_url="https://output.example.com/job-s2v",
                transaction_id=None,
                submitted_at=None,
            )

        def get_job_categories(self, limit=100):
            from unittest.mock import MagicMock
            cats = MagicMock()
            c = MagicMock()
            c.id = "cat-1"
            c.name = "Qiskit"
            cats.categories = [c]
            return cats

        def get_job_subcategories(self, category_id, limit=100):
            from unittest.mock import MagicMock
            subs = MagicMock()
            s = MagicMock()
            s.id = "sub-1"
            s.name = "sub-1"
            subs.categories = [s]
            return subs

    backend._scheduler = _SchedulerStub()
    backend._config = {
        "organization_id": "org-1",
        "backend_class_id": "bk-1",
        "job_subcategory_id": "sub-1",
    }

    counts = {"00": 70, "11": 30}
    requests_mock.get("https://output.example.com/job-s2v", text=json.dumps(counts))

    qc = QuantumCircuit(2)
    qc.h(0)
    qc.cx(0, 1)
    qc.measure_all()

    sampler = BackendSamplerV2(backend=backend)
    res = sampler.run([(qc, None, 100)]).result()
    assert len(res) == 1
    total = sum(res[0].data.meas.get_counts().values())
    assert total == 100


def test_backend_requires_scheduler():
    """Test that Backend.run raises RuntimeError if scheduler is missing."""
    cap = _caps_2q()
    qc = QuantumCircuit(2)
    qc.h(0)
    qc.measure_all()

    backend = OpenQuantumBackend(name="test", capabilities=cap, scheduler=None, config=None)
    with pytest.raises(RuntimeError, match="provide scheduler and config"):
        backend.run([qc])

    backend = OpenQuantumBackend(
        name="test",
        capabilities=cap,
        scheduler=None,
        config={"backend_class_id": "test", "job_subcategory_id": "test"}
    )
    with pytest.raises(RuntimeError, match="provide scheduler and config"):
        backend.run([qc])


def test_backend_requires_config_keys():
    """Test that Backend.run raises ValueError if required config keys are missing."""
    cap = _caps_2q()
    qc = QuantumCircuit(2)
    qc.h(0)
    qc.measure_all()

    mock_scheduler = object()

    backend = OpenQuantumBackend(name="test", capabilities=cap, scheduler=mock_scheduler, config={})
    with pytest.raises(ValueError, match="Missing config keys.*backend_class_id.*job_subcategory_id"):
        backend.run([qc])

    backend = OpenQuantumBackend(
        name="test",
        capabilities=cap,
        scheduler=mock_scheduler,
        config={"backend_class_id": "test:backend"}
    )
    with pytest.raises(ValueError, match="Missing config keys.*job_subcategory_id"):
        backend.run([qc])

    backend = OpenQuantumBackend(
        name="test",
        capabilities=cap,
        scheduler=mock_scheduler,
        config={"job_subcategory_id": "test:subcategory"}
    )
    with pytest.raises(ValueError, match="Missing config keys.*backend_class_id"):
        backend.run([qc])


def test_prx_gate_define_does_not_recurse():
    """Test that PRXGate._define() doesn't cause infinite recursion."""
    import numpy as np

    caps = {
        "n_qubits": 2,
        "native_ops": [
            {"name": "prx", "arity": 1, "params": ["theta", "phi"]},
            {"name": "cz", "arity": 2},
            {"name": "measure", "arity": 1},
        ],
        "topology": {"directed_edges": False, "coupling_map": [[0, 1]]},
    }

    target = build_target_from_capabilities(caps)
    assert "prx" in target.operation_names

    PRXGateClass = None
    for inst in target.operations:
        if inst.name == "prx":
            PRXGateClass = type(inst)
            break

    assert PRXGateClass is not None, "PRXGate not found in target"

    gate = PRXGateClass(np.pi / 2, np.pi / 4)

    defn = gate.definition
    assert defn is not None
    assert defn.num_qubits == 1
    assert len(defn.data) == 3


def test_prx_gate_transpilation_no_recursion():
    """Test that transpiling a circuit with PRX gates doesn't cause recursion."""
    from qiskit import transpile

    caps = {
        "n_qubits": 2,
        "native_ops": [
            {"name": "prx", "arity": 1, "params": ["theta", "phi"]},
            {"name": "cz", "arity": 2},
            {"name": "measure", "arity": 1},
        ],
        "topology": {"directed_edges": False, "coupling_map": [[0, 1]]},
    }

    backend = OpenQuantumBackend(name="test-prx", capabilities=caps)

    qc = QuantumCircuit(2)
    qc.h(0)
    qc.cx(0, 1)
    qc.measure_all()

    tqc = transpile(qc, backend=backend, optimization_level=1)
    assert tqc is not None
    assert tqc.num_qubits == 2



from __future__ import annotations

from typing import Any


def export_circuit_to_qasm(circuit: Any, fmt: str) -> bytes:
    f = (fmt or "qasm3").lower()
    if f == "qasm3":
        try:
            import qiskit.qasm3 as qasm3
        except Exception as e:
            raise RuntimeError(
                "Qiskit qasm3 exporter is required. Install: pip install 'openquantum-sdk[qiskit]'"
            ) from e
        qasm = qasm3.dumps(circuit)
        return qasm.encode("utf-8")
    if f == "qasm2":
        try:
            from qiskit import qasm2

            qasm = qasm2.dumps(circuit)
            return qasm.encode("utf-8")
        except Exception:
            try:
                qasm = circuit.qasm()
                return qasm.encode("utf-8")
            except Exception as e:
                raise RuntimeError("Failed to export circuit to OpenQASM 2.") from e
    raise ValueError("export_format must be 'qasm2' or 'qasm3'")



from __future__ import annotations
from openquantum_sdk.models import JobRead
from openquantum_sdk.clients import SchedulerClient
from openquantum_sdk_qiskit.oq_sampler import OQSampler
from qiskit import QuantumCircuit

import json
from typing import Dict

import pytest

qiskit = pytest.importorskip("qiskit")


def _simple_bell(shots: int = 100) -> QuantumCircuit:
    qc = QuantumCircuit(2)
    qc.metadata = {"foo": "fooo"}
    qc.h(0)
    qc.cx(0, 1)
    qc.measure_all()
    return qc


class _SchedulerStub(SchedulerClient):
    base_url = "https://scheduler.openquantum.com"

    def __init__(self, output_url: str, counts: Dict[str, int]):
        self._output_url = output_url
        self._counts = counts

    def submit_job(self, config, file_content=None, file_path=None):
        return JobRead(
            id="job-pub",
            status="Completed",
            input_data_url="https://input.example.com/job-pub",
            job_preparation_id="prep",
            execution_plan_id="plan",
            queue_priority_id="prio",
            message=None,
            output_data_url=self._output_url,
            transaction_id=None,
            submitted_at=None,
        )

    def get_job_categories(self, limit=100):
        from unittest.mock import MagicMock
        cats = MagicMock()
        c = MagicMock()
        c.id = "cat-1"
        c.name = "Qiskit"
        cats.categories = [c]
        return cats

    def get_job_subcategories(self, category_id, limit=100):
        from unittest.mock import MagicMock
        subs = MagicMock()
        s = MagicMock()
        s.id = "sub-1"
        s.name = "sub-1"
        subs.categories = [s]
        return subs


def test_sampler_v2_maps_counts_to_pub_databin(requests_mock):
    qc = _simple_bell(100)
    requests_mock.get("https://scheduler.openquantum.com/v1/jobs/categories?limit=200", json={
        "job_categories": [{"id": "cat-1", "name": "Qiskit", "job_subcategories": [{"id": "sub-1", "name": "sub-1"}]}]
    })

    output_url = "https://output.example.com/job-pub"
    counts = {"0": 50, "1": 50}
    requests_mock.get(output_url, text=json.dumps(counts))

    sampler = OQSampler(
        backend=object(),
        scheduler=_SchedulerStub(output_url, counts),
        config={
            "organization_id": "org-1",
            "backend_class_id": "bk-1",
            "job_subcategory_id": "sub-1",
        },
        export_format="qasm3",
    )

    job = sampler.run([(qc, None, 100)])
    res = job.result()
    assert len(res) == 1
    pub0 = res[0]

    total = sum(pub0.data.meas.get_counts().values())
    assert total == 100
    assert pub0.metadata["shots"] == 100
    assert pub0.metadata["circuit_metadata"] == qc.metadata


def test_sampler_v2_validates_export_format_on_init_and_run():
    with pytest.raises(ValueError):
        OQSampler(backend=object(), scheduler=None, config=None, export_format="invalid")


def test_sampler_v2_requires_scheduler():
    """Test that Sampler raises RuntimeError if scheduler is missing."""
    qc = _simple_bell(10)
    sampler = OQSampler(backend=object(), scheduler=None, config=None, export_format="qasm3")
    with pytest.raises(RuntimeError, match="provide scheduler and config"):
        _ = sampler.run([(qc, None, 10)])


def test_sampler_v2_requires_config_keys():
    """Test that Sampler raises ValueError if required config keys are missing."""
    qc = _simple_bell(10)
    mock_scheduler = object()

    sampler = OQSampler(backend=object(), scheduler=mock_scheduler, config={}, export_format="qasm3")
    with pytest.raises(ValueError, match="Missing config keys.*backend_class_id.*job_subcategory_id"):
        _ = sampler.run([(qc, None, 10)])

    sampler = OQSampler(
        backend=object(),
        scheduler=mock_scheduler,
        config={"backend_class_id": "test:backend"},
        export_format="qasm3"
    )
    with pytest.raises(ValueError, match="Missing config keys.*job_subcategory_id"):
        _ = sampler.run([(qc, None, 10)])

    sampler = OQSampler(
        backend=object(),
        scheduler=mock_scheduler,
        config={"job_subcategory_id": "test:subcategory"},
        export_format="qasm3"
    )
    with pytest.raises(ValueError, match="Missing config keys.*backend_class_id"):
        _ = sampler.run([(qc, None, 10)])


def test_sampler_v2_uses_run_time_export_format_over_default(requests_mock):
    requests_mock.get("https://scheduler.openquantum.com/v1/jobs/categories?limit=200", json={
        "job_categories": [{"id": "cat-1", "name": "Qiskit", "job_subcategories": [{"id": "sub-1", "name": "sub-1"}]}]
    })

    output_url = "https://output.example.com/job-fmt"
    counts = {"0": 1}

    class _Scheduler(SchedulerClient):
        base_url = "https://scheduler.openquantum.com"

        def submit_job(self, config, file_content=None, file_path=None):
            return JobRead(
                id="job-fmt",
                status="Completed",
                input_data_url="",
                job_preparation_id="",
                execution_plan_id="",
                queue_priority_id="",
                message=None,
                output_data_url=output_url,
                transaction_id=None,
                submitted_at=None,
            )

        def get_job_categories(self, limit=100):
            from unittest.mock import MagicMock
            cats = MagicMock()
            c = MagicMock()
            c.id = "cat-1"
            c.name = "Qiskit"
            cats.categories = [c]
            return cats

        def get_job_subcategories(self, category_id, limit=100):
            from unittest.mock import MagicMock
            subs = MagicMock()
            s = MagicMock()
            s.id = "sub-1"
            s.name = "sub-1"
            subs.categories = [s]
            return subs

    import requests_mock as rqm

    with rqm.Mocker() as m:
        m.get(output_url, text=json.dumps(counts))
        m.get("https://scheduler.openquantum.com/v1/jobs/categories?limit=200", json={
            "job_categories": [{"id": "cat-1", "name": "Qiskit", "job_subcategories": [{"id": "sub-1", "name": "sub-1"}]}]
        })
        sampler = OQSampler(
            backend=object(),
            scheduler=_Scheduler(),
            config={
                "organization_id": "org-1",
                "backend_class_id": "bk-1",
                "job_subcategory_id": "sub-1",
            },
            export_format="qasm3",
        )
        qc = _simple_bell(1)
        job = sampler.run([(qc, None, 1)], export_format="qasm2")
        assert job.metadata["export_format"] == "qasm2"
        _ = job.result()


def test_sampler_bitarray_uses_num_clbits_not_shots(requests_mock):
    """Test that BitArray.from_samples uses circuit.num_clbits, not shots.

    This prevents the bug where histograms show 100+ leading zeros when
    shots > num_clbits.
    """
    requests_mock.get("https://scheduler.openquantum.com/v1/jobs/categories?limit=200", json={
        "job_categories": []
    })

    output_url = "https://output.example.com/job-bits"
    counts = {"00": 500, "11": 500}

    class _Scheduler(SchedulerClient):
        base_url = "https://scheduler.openquantum.com"

        def submit_job(self, config, file_content=None, file_path=None):
            return JobRead(
                id="job-bits",
                status="Completed",
                input_data_url="https://input.example.com/job-bits",
                job_preparation_id="prep-bits",
                execution_plan_id="plan",
                queue_priority_id="prio",
                message=None,
                output_data_url=output_url,
                transaction_id=None,
                submitted_at=None,
            )

        def download_job_output(self, job):
            return counts

    qc = QuantumCircuit(2, 2)
    qc.h(0)
    qc.cx(0, 1)
    qc.measure([0, 1], [0, 1])

    sampler = OQSampler(
        backend=object(),
        scheduler=_Scheduler(output_url, counts),
        config={
            "organization_id": "org-1",
            "backend_class_id": "bk-1",
            "job_subcategory_id": "sub-1",
        },
        export_format="qasm3",
    )

    job = sampler.run([(qc, None, 1000)])
    result = job.result()

    result_counts = result[0].data.meas.get_counts()

    for key in result_counts.keys():
        assert len(key) == 2, f"Expected 2-bit keys, got {len(key)}-bit key: {key}"



import pytest
from unittest.mock import MagicMock
from qiskit import QuantumCircuit
from qiskit.quantum_info import SparsePauliOp
from qiskit.providers import BackendV2
from openquantum_sdk_qiskit.oq_backend import OpenQuantumBackend
from openquantum_sdk_qiskit import OQEstimator


def test_estimator_initialization():
    backend = MagicMock(spec=BackendV2)
    backend.name = "mock_backend"
    estimator = OQEstimator(backend=backend)
    assert estimator.backend == backend
    assert estimator.options is not None


def test_estimator_run_no_scheduler():
    backend = OpenQuantumBackend(name="Rigetti Ankaa-3")
    estimator = OQEstimator(backend=backend)
    qc = QuantumCircuit(1)
    qc.h(0)
    qc.measure_all()
    op = SparsePauliOp("Z")

    with pytest.raises(RuntimeError, match="Scheduler submission is required"):
        estimator.run([(qc, op)]).result()



from __future__ import annotations
from openquantum_sdk.qiskit import OpenQuantumService
from qiskit.primitives import BackendSamplerV2
from qiskit import QuantumCircuit
import json
from typing import Dict
import pytest
qiskit = pytest.importorskip("qiskit")


def _caps_2q() -> Dict:
    return {
        "n_qubits": 2,
        "timing": {"dt": 2.0e-9, "durations": [{"op": "cx", "qubits": [0, 1], "value_s": 5.0e-7}]},
        "native_ops": [
            {"name": "x", "arity": 1},
            {"name": "h", "arity": 1},
            {"name": "cx", "arity": 2},
        ],
        "topology": {"directed_edges": False, "coupling_map": [[0, 1]]},
        "noise": {"gate_error": [{"op": "cx", "qubits": [0, 1], "prob": 0.01}]},
        "limits": {"max_circuits": 100},
    }


def test_return_backend_and_run_counts_to_memory(requests_mock):
    svc = OpenQuantumService()
    backend = svc.return_backend(
        name="oq://test/backend",
        capabilities_source=_caps_2q(),
        config={
            "organization_id": "org-1",
            "backend_class_id": "bk-1",
            "job_subcategory_id": "sub-1",
        },
        export_format="qasm3",
    )

    counts = {"00": 60, "11": 40}
    output_url = "https://output.example.com/job-bkrun"

    from openquantum_sdk.clients import SchedulerClient
    from openquantum_sdk.models import JobRead
    from unittest.mock import MagicMock

    class _SchedulerStub(SchedulerClient):
        base_url = "https://scheduler.openquantum.com"

        def submit_job(self, config, file_content=None, file_path=None):
            return JobRead(
                id="job-bkrun",
                status="Completed",
                input_data_url="",
                job_preparation_id="",
                execution_plan_id="",
                queue_priority_id="",
                message=None,
                output_data_url=output_url,
                transaction_id=None,
                submitted_at=None,
            )

        def get_job_categories(self, limit=100):
            cats = MagicMock()
            c = MagicMock()
            c.id = "cat-1"
            c.name = "Qiskit"
            cats.categories = [c]
            return cats

        def get_job_subcategories(self, category_id, limit=100):
            subs = MagicMock()
            s = MagicMock()
            s.id = "sub-1"
            s.name = "sub-1"
            subs.categories = [s]
            return subs

    backend._scheduler = _SchedulerStub()

    requests_mock.get("https://scheduler.openquantum.com/v1/jobs/categories?limit=200", json={
        "job_categories": [{"id": "cat-1", "name": "Qiskit", "job_subcategories": [{"id": "sub-1", "name": "sub-1"}]}]
    })

    requests_mock.get(output_url, text=json.dumps(counts))

    qc = QuantumCircuit(2)
    qc.h(0)
    qc.cx(0, 1)
    qc.measure_all()

    job = backend.run([qc], shots=100, memory=True)
    result = job.result()
    assert len(result.results) == 1
    assert result.results[0].shots == 100
    assert isinstance(result.results[0].data.memory, list)
    assert len(result.results[0].data.memory) == 100


def test_create_sampler_wires_scheduler_and_runs(requests_mock):
    svc = OpenQuantumService()
    backend = svc.return_backend(
        name="oq://test/backend",
        capabilities_source=_caps_2q(),
        config={
            "organization_id": "org-1",
            "backend_class_id": "bk-1",
            "job_subcategory_id": "sub-1",
        },
    )

    from openquantum_sdk.clients import SchedulerClient
    from openquantum_sdk.models import JobRead

    output_url = "https://output.example.com/job-sampler"
    counts = {"00": 70, "11": 30}
    requests_mock.get(output_url, text=json.dumps(counts))

    class _SchedulerStub(SchedulerClient):
        base_url = "https://scheduler.openquantum.com"

        def submit_job(self, config, file_content=None, file_path=None):
            return JobRead(
                id="job-sampler",
                status="Completed",
                input_data_url="",
                job_preparation_id="",
                execution_plan_id="",
                queue_priority_id="",
                message=None,
                output_data_url=output_url,
                transaction_id=None,
                submitted_at=None,
            )

        def get_job_categories(self, limit=100):
            from unittest.mock import MagicMock
            cats = MagicMock()
            c = MagicMock()
            c.id = "cat-1"
            c.name = "Qiskit"
            s = MagicMock()
            s.id = "sub-1"
            s.name = "Sampler"

            cats.categories = [c]
            return cats

        def get_job_subcategories(self, category_id, limit=100):
            from unittest.mock import MagicMock
            subs = MagicMock()
            s = MagicMock()
            s.id = "sub-1"
            s.name = "sub-1"
            subs.categories = [s]
            return subs

    svc.scheduler = _SchedulerStub()
    backend._scheduler = svc.scheduler

    requests_mock.get("https://scheduler.openquantum.com/v1/jobs/categories?limit=200", json={
        "job_categories": [{"id": "cat-1", "name": "Qiskit", "job_subcategories": [{"id": "sub-1", "name": "Sampler"}]}]
    })

    sampler = svc.create_sampler(
        backend=backend,
        organization_id="org-1",
        backend_class_id="bk-1",
        job_subcategory_id="sub-1",
        export_format="qasm3",
    )

    qc = QuantumCircuit(2)
    qc.h(0)
    qc.cx(0, 1)
    qc.measure_all()

    b_sampler = BackendSamplerV2(backend=backend)
    res_backend = b_sampler.run([(qc, None, 100)]).result()
    assert sum(res_backend[0].data.meas.get_counts().values()) == 100

    res = sampler.run([(qc, None, 100)]).result()
    assert sum(res[0].data.meas.get_counts().values()) == 100
